#include "hvif-light.h"

#include <assert.h>
#include <cairo.h>
#include <math.h>
#include <stdint.h>
#include <stdlib.h>

#define INTERNAL_DATASTRUCTURES
#include "hvif-light.c"
#undef INTERNAL_DATASTRUCTURES

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

/* TODO:
 *  - Color space
 *  - Contour
 *  - LOD
 *  - Check miter limit
 */

void
cairo_matrix_init_from_matrix(
	cairo_matrix_t* matrix, const hvif_matrix* hvif_matrix)
{
	cairo_matrix_init(
		matrix, hvif_matrix->xx, hvif_matrix->yx, hvif_matrix->xy, hvif_matrix->yy,
		hvif_matrix->x0, hvif_matrix->y0);
}

void
print_matrix(const hvif_matrix* matrix)
{
	printf("x_new = %f * x + %f * y + %f\n", matrix->xx, matrix->xy, matrix->x0);
	printf("y_new = %f * x + %f * y + %f\n", matrix->yx, matrix->yy, matrix->y0);
}

#define R(c) (COLOR_GET_RED(c) / 255.0)
#define G(c) (COLOR_GET_GREEN(c) / 255.0)
#define B(c) (COLOR_GET_BLUE(c) / 255.0)
#define A(c) (COLOR_GET_ALPHA(c) / 255.0)

void
create_diamond_patch(
	cairo_pattern_t* pat, double s1, double s2, hvif_color c1, hvif_color c2)
{
	cairo_mesh_pattern_begin_patch(pat);
	cairo_mesh_pattern_move_to(pat, s1, s1);
	cairo_mesh_pattern_line_to(pat, s2, s2);
	cairo_mesh_pattern_line_to(pat, s2, -s2);
	cairo_mesh_pattern_line_to(pat, s1, -s1);
	cairo_mesh_pattern_set_corner_color_rgba(pat, 0, R(c1), G(c1), B(c1), A(c1));
	cairo_mesh_pattern_set_corner_color_rgba(pat, 1, R(c2), G(c2), B(c2), A(c2));
	cairo_mesh_pattern_set_corner_color_rgba(pat, 2, R(c2), G(c2), B(c2), A(c2));
	cairo_mesh_pattern_set_corner_color_rgba(pat, 3, R(c1), G(c1), B(c1), A(c1));
	cairo_mesh_pattern_end_patch(pat);
	cairo_mesh_pattern_begin_patch(pat);
	cairo_mesh_pattern_move_to(pat, s1, -s1);
	cairo_mesh_pattern_line_to(pat, s2, -s2);
	cairo_mesh_pattern_line_to(pat, -s2, -s2);
	cairo_mesh_pattern_line_to(pat, -s1, -s1);
	cairo_mesh_pattern_set_corner_color_rgba(pat, 0, R(c1), G(c1), B(c1), A(c1));
	cairo_mesh_pattern_set_corner_color_rgba(pat, 1, R(c2), G(c2), B(c2), A(c2));
	cairo_mesh_pattern_set_corner_color_rgba(pat, 2, R(c2), G(c2), B(c2), A(c2));
	cairo_mesh_pattern_set_corner_color_rgba(pat, 3, R(c1), G(c1), B(c1), A(c1));
	cairo_mesh_pattern_end_patch(pat);
	cairo_mesh_pattern_begin_patch(pat);
	cairo_mesh_pattern_move_to(pat, -s1, -s1);
	cairo_mesh_pattern_line_to(pat, -s2, -s2);
	cairo_mesh_pattern_line_to(pat, -s2, s2);
	cairo_mesh_pattern_line_to(pat, -s1, s1);
	cairo_mesh_pattern_set_corner_color_rgba(pat, 0, R(c1), G(c1), B(c1), A(c1));
	cairo_mesh_pattern_set_corner_color_rgba(pat, 1, R(c2), G(c2), B(c2), A(c2));
	cairo_mesh_pattern_set_corner_color_rgba(pat, 2, R(c2), G(c2), B(c2), A(c2));
	cairo_mesh_pattern_set_corner_color_rgba(pat, 3, R(c1), G(c1), B(c1), A(c1));
	cairo_mesh_pattern_end_patch(pat);
	cairo_mesh_pattern_begin_patch(pat);
	cairo_mesh_pattern_move_to(pat, -s1, s1);
	cairo_mesh_pattern_line_to(pat, -s2, s2);
	cairo_mesh_pattern_line_to(pat, s2, s2);
	cairo_mesh_pattern_line_to(pat, s1, s1);
	cairo_mesh_pattern_set_corner_color_rgba(pat, 0, R(c1), G(c1), B(c1), A(c1));
	cairo_mesh_pattern_set_corner_color_rgba(pat, 1, R(c2), G(c2), B(c2), A(c2));
	cairo_mesh_pattern_set_corner_color_rgba(pat, 2, R(c2), G(c2), B(c2), A(c2));
	cairo_mesh_pattern_set_corner_color_rgba(pat, 3, R(c1), G(c1), B(c1), A(c1));
	cairo_mesh_pattern_end_patch(pat);
}

void
create_conic_patch(
	cairo_pattern_t* pat, double r, double angle1, double angle2, hvif_color c1,
	hvif_color c2)
{
	double sin1 = r * sin(angle1);
	double cos1 = r * cos(angle1);
	double sin2 = r * sin(angle2);
	double cos2 = r * cos(angle2);
	double h = 4.0 / 3.0 * tan((angle2 - angle1) / 4.0);
	cairo_mesh_pattern_begin_patch(pat);
	cairo_mesh_pattern_move_to(pat, 0, 0);
	cairo_mesh_pattern_line_to(pat, cos1, sin1);
	cairo_mesh_pattern_curve_to(
		pat, cos1 - h * sin1, sin1 + h * cos1, cos2 + h * sin2, sin2 - h * cos2,
		cos2, sin2);
	cairo_mesh_pattern_line_to(pat, 0, 0);
	cairo_mesh_pattern_set_corner_color_rgba(pat, 0, R(c1), G(c1), B(c1), A(c1));
	cairo_mesh_pattern_set_corner_color_rgba(pat, 1, R(c1), G(c1), B(c1), A(c1));
	cairo_mesh_pattern_set_corner_color_rgba(pat, 2, R(c2), G(c2), B(c2), A(c2));
	cairo_mesh_pattern_set_corner_color_rgba(pat, 3, R(c2), G(c2), B(c2), A(c2));
	cairo_mesh_pattern_end_patch(pat);
}

void
create_pattern_style(cairo_pattern_t** pat, hvif_style* style)
{
	assert(style->num_stops > 0);

	double width = 64.0;
	double height = 64.0;
	cairo_matrix_t transformation;
	cairo_matrix_init_from_matrix(&transformation, &style->transformation);
	cairo_matrix_transform_point(&transformation, &width, &height);
	double extremum = width > height ? width : height;

	switch (style->gradient_type) {
		case GRADIENT_LINEAR:
			*pat = cairo_pattern_create_linear(-64, 32, 64, 32);
			for (unsigned i = 0; i < style->num_stops; ++i) {
				hvif_color c = style->colors[i];
				cairo_pattern_add_color_stop_rgba(
					*pat, style->offsets[i], R(c), G(c), B(c), A(c));
			}
			break;
		case GRADIENT_CIRCULAR:
			*pat = cairo_pattern_create_radial(0, 0, 0, 0, 0, 64);
			for (unsigned i = 0; i < style->num_stops; ++i) {
				hvif_color c = style->colors[i];
				cairo_pattern_add_color_stop_rgba(
					*pat, style->offsets[i], R(c), G(c), B(c), A(c));
			}
			break;
		case GRADIENT_DIAMOND:
			*pat = cairo_pattern_create_mesh();
			if (style->offsets[0] != 0.0) {
				create_diamond_patch(
					*pat, 0, 64.0 * style->offsets[0], style->colors[0],
					style->colors[0]);
			}
			for (unsigned i = 0; i + 1 < style->num_stops; ++i) {
				double s1 = style->offsets[i] * 64.0;
				double s2 = style->offsets[i + 1] * 64.0;
				hvif_color c1 = style->colors[i];
				hvif_color c2 = style->colors[i + 1];
				create_diamond_patch(*pat, s1, s2, c1, c2);
			}
			unsigned i = style->num_stops - 1;
			create_diamond_patch(
				*pat, 64.0 * style->offsets[i], extremum, style->colors[i],
				style->colors[i]);
			break;
		case GRADIENT_CONIC:
			*pat = cairo_pattern_create_mesh();
			if (style->offsets[0] != 0.0) {
				create_conic_patch(
					*pat, extremum, 0, M_PI * style->offsets[0], style->colors[0],
					style->colors[0]);
				create_conic_patch(
					*pat, extremum, 0, -M_PI * style->offsets[0], style->colors[0],
					style->colors[0]);
			}
			for (unsigned i = 0; i + 1 < style->num_stops; ++i) {
				create_conic_patch(
					*pat, extremum, M_PI * style->offsets[i],
					M_PI * style->offsets[i + 1], style->colors[i], style->colors[i + 1]);
				create_conic_patch(
					*pat, extremum, -M_PI * style->offsets[i],
					-M_PI * style->offsets[i + 1], style->colors[i],
					style->colors[i + 1]);
			}
			i = style->num_stops - 1;
			if (style->offsets[i] != 1.0) {
				create_conic_patch(
					*pat, extremum, M_PI * style->offsets[i], 0, style->colors[i],
					style->colors[i]);
				create_conic_patch(
					*pat, extremum, -M_PI * style->offsets[i], 0, style->colors[i],
					style->colors[i]);
			}
			break;
		case GRADIENT_XY: printf("GRADIENT_XY not supported\n"); exit(1);
		case GRADIENT_SQRT_XY: printf("GRADIENT_SQRT_XY not supported\n"); exit(1);
		default: printf("unsupported gradient type\n"); exit(1);
	}

	if (cairo_matrix_invert(&transformation) != CAIRO_STATUS_SUCCESS) {
		printf("could not invert style transformation\n");
		exit(1);
	} else
		cairo_pattern_set_matrix(*pat, &transformation);
}

void
create_path(cairo_t* cr, cairo_path_t** path, hvif_path* hvif_path)
{
	assert(hvif_path->num_points > 0);

	cairo_new_path(cr);
	cairo_move_to(cr, hvif_path->points[0].x, hvif_path->points[0].y);
	for (unsigned i = 1; i < hvif_path->num_points; ++i) {
		hvif_point cp1 = hvif_path->points[3 * i - 1];
		hvif_point cp2 = hvif_path->points[3 * i + 1];
		hvif_point p = hvif_path->points[3 * i];
		cairo_curve_to(cr, cp1.x, cp1.y, cp2.x, cp2.y, p.x, p.y);
	}

	if (hvif_path->closed) {
		unsigned max = hvif_path->num_points - 1;
		hvif_point cp1 = hvif_path->points[3 * max + 2];
		hvif_point cp2 = hvif_path->points[1];
		hvif_point p = hvif_path->points[0];
		cairo_curve_to(cr, cp1.x, cp1.y, cp2.x, cp2.y, p.x, p.y);
	}
	cairo_close_path(cr);
	*path = cairo_copy_path_flat(cr);
}

void
render_shape(
	cairo_t* cr, hvif_shape* shape, cairo_pattern_t** pattern,
	cairo_path_t** path)
{
	cairo_matrix_t old_transform;
	cairo_get_matrix(cr, &old_transform);
	cairo_matrix_t shape_transform;
	cairo_matrix_init_from_matrix(&shape_transform, &shape->transformation);
	cairo_transform(cr, &shape_transform);

	int stroke_idx = -1;
	for (unsigned i = 0; i < shape->num_transformers; ++i) {
		hvif_transformer* t = &shape->transformers[i];
		switch (t->transformer_type) {
			case TRANSFORMER_TYPE_STROKE:
				if (stroke_idx != -1)
					printf("unsupported: multiple stroke transform\n");
				stroke_idx = i;
				break;
			case TRANSFORMER_TYPE_AFFINE:
				/* TODO: this is a guess what this does, but it's not exposed in
				         Icon-o-Matic. */
				printf("warning: affine transform\n");
				cairo_matrix_init_from_matrix(
					&shape_transform, &t->transformer.affine.matrix);
				break;
			default: printf("unsupported: transform\n");
		}
	}

	cairo_new_path(cr);
	for (unsigned i = 0; i < shape->num_paths; ++i) {
		cairo_append_path(cr, path[shape->path_idxs[i]]);
		cairo_new_sub_path(cr);
	}
	cairo_set_source(cr, pattern[shape->style_idx]);

	if (stroke_idx >= 0) {
		hvif_transformer* t = &shape->transformers[stroke_idx];
		cairo_set_line_width(cr, t->transformer.stroke.width);
		switch (t->transformer.stroke.line_join) {
			case (MITER_JOIN): cairo_set_line_join(cr, CAIRO_LINE_JOIN_MITER); break;
			case (ROUND_CAP): cairo_set_line_join(cr, CAIRO_LINE_JOIN_ROUND); break;
			case (BEVEL_JOIN): cairo_set_line_join(cr, CAIRO_LINE_JOIN_BEVEL); break;
			default: printf("unsupported: line join\n");
		}
		cairo_set_miter_limit(cr, t->transformer.stroke.miter_limit);
		switch (t->transformer.stroke.line_cap) {
			case (BUTT_CAP): cairo_set_line_cap(cr, CAIRO_LINE_CAP_BUTT); break;
			case (SQUARE_CAP): cairo_set_line_cap(cr, CAIRO_LINE_CAP_SQUARE); break;
			case (ROUND_CAP): cairo_set_line_cap(cr, CAIRO_LINE_CAP_ROUND); break;
			default: printf("illegal line cap"); exit(1);
		}
		cairo_stroke(cr);
	} else
		cairo_fill(cr);

	cairo_set_matrix(cr, &old_transform);

	if (shape->hinting) printf("unsupported: hinting\n");
	/* TODO: implement LOD */
}

bool
hvif_render_image(const char* filename, hvif_image* image)
{
	unsigned image_size = 512;

	cairo_surface_t* surface =
		cairo_image_surface_create(CAIRO_FORMAT_ARGB32, image_size, image_size);
	cairo_t* cr = cairo_create(surface);
	cairo_scale(cr, image_size / 64.0, image_size / 64.0);

	cairo_pattern_t** patterns =
		malloc(sizeof(cairo_pattern_t*) * image->num_styles);
	if (!patterns) return false;
	cairo_path_t** paths = malloc(sizeof(cairo_path_t*) * image->num_paths);
	if (!paths) return false;

	for (unsigned i = 0; i < image->num_styles; ++i) {
		create_pattern_style(&patterns[i], &image->styles[i]);
		/* TODO: error handling */
	}
	for (unsigned i = 0; i < image->num_paths; ++i) {
		create_path(cr, &paths[i], &image->paths[i]);
		/* TODO: error handling */
	}

	for (unsigned i = 0; i < image->num_shapes; ++i) {
		render_shape(cr, &image->shapes[i], patterns, paths);
	}

	for (unsigned i = 0; i < image->num_paths; ++i) {
		cairo_path_destroy(paths[i]);
	}
	for (unsigned i = 0; i < image->num_styles; ++i) {
		cairo_pattern_destroy(patterns[i]);
	}
	bool result = true;
	if (cairo_surface_write_to_png(surface, filename) != CAIRO_STATUS_SUCCESS)
		result = false;
	cairo_surface_destroy(surface);
	cairo_destroy(cr);
	return result;
}
